from typing import Tuple, List, Union

from parsee.utils.enums import DocumentType
from parsee.extraction.extractor_elements import ExtractedEl


class RawToJsonConverter:

    def __init__(self, doc_type: DocumentType):
        self.doc_type = doc_type

    def convert(self, file_path: str) -> List[ExtractedEl]:
        raise NotImplemented



from typing import *
import re
from decimal import Decimal
import time
import hashlib

import numpy as np


years_abs_strings = []
current_year = int(time.strftime("%Y"))
for yyyy in range(1980, current_year + 3):
    years_abs_strings.append(str(yyyy))


def clean_numeric_value_llm(string_val) -> Union[float, None]:
    matches = re.findall(r'(\d+(\d|,|\.|)*\d|\d)', string_val)
    last_numeric = None
    for match, _ in matches:
        if match.strip() != "":
            cleaned = clean_numeric_value(match)
            if cleaned is not None:
                last_numeric = float(cleaned)
    return last_numeric


def get_entity_value(string_val) -> Union[str, None]:
    match = re.search(r'(.+)( is:? | are:? |: )(.+)\.?', string_val)
    if match:
        return match.group(3)
    else:
        return None


def get_date_regex(string_val) -> Union[str, None]:
    match = re.search(r'(\d\d\d\d\-\d\d\-\d\d)', string_val)
    if match:
        return match.group(1)
    else:
        return None


def is_negative(cell_str):
    # minus
    if re.search(r'(-|—|–|‒|―|–|−)( | |)*\d', cell_str.strip()):
        return True
    # brackets
    if re.search(r'\([\d ,.%]+(\)|\b)', cell_str.strip()):
        return True
    return False


def comma_separator_thousands(cell_str):
    if re.search(r'\b[0-9]{1,3}[,][0-9]{3}\b', cell_str):
        return True
    return False


def dot_separator_thousands(cell_str):
    if re.search(r'\b[0-9]{1,3}[.][0-9]{3}\b', cell_str):
        return True
    return False


def clean_numeric_value(cell_str: str):
    if cell_str.strip() == "":
        return None
    mult = 1
    if is_negative(cell_str):
        mult = -1

    cell_str = re.sub(r'[^0-9,.]', '', cell_str)

    # clean thousands separator
    if comma_separator_thousands(cell_str):
        cell_str = re.sub(r',', "", cell_str)
    elif dot_separator_thousands(cell_str):
        cell_str = re.sub(r'\.', "", cell_str)

    # now also replace the comma with a dot should it be used in any case
    cell_str = re.sub(r',', ".", cell_str)

    if cell_str.replace('.', '', 1).isdigit():
        return Decimal(cell_str) * mult
    else:
        return None


def is_number_cell(cell_str):
    cell_str = str(cell_str)
    if cell_str is None:
        return False
    cell_str = re.sub(r'(\([^0-9 ]*\))|[^0-9A-Za-z/]', '', cell_str)
    if cell_str.isdigit():
        return True
    else:
        return False


def words_contained(cell_str, lower=False) -> List[str]:
    if lower:
        cell_str = cell_str.lower()
    return list(filter(lambda x: x != "", re.sub('[^A-Za-z0-9%$€£¥]', ' ', cell_str).split(" ")))


# cleans text for word embeddings
def clean_text_for_word_vectors2(text, base_year=None, remove_special_chars=False, remove_all_numbers=False, number_token=" xnumberx "):

    if text is None:
        return ""

    text = text.lower()

    # replace new lines with space
    text = " ".join(text.splitlines())

    # replace numbers with commas or dots in them with a token
    text = re.sub(r'(\b([0-9]{1,3}[,.])+[0-9]+\b|\b[0-9]+[,.][0-9]+\b)', number_token, text)

    if remove_special_chars:
        # remove all special characters apart from a few selected ones
        text = re.sub(r'[^\w\d\s,.;:#\'\"+*!$€%&/\[\]()-_=<>§`´]', " ", text)

    # insert a space around each special character
    text = re.sub(r'([^0-9A-Za-z ])', ' \\1 ', text)

    if base_year is not None:
        base_year = int(base_year)
        base_year_offsets = [-3, -2, -1, 0, 1, 2, 3]
        # replace years
        for offset in base_year_offsets:
            text = text.replace(str(base_year + offset),
                                " xbaseyear" + ("m" if offset <= 0 else "p") + str(abs(offset)) + "x ")

    if remove_all_numbers:
        text = re.sub(r'(\b[0-9]+\b)', number_token, text)

    # replace double space with single
    text = re.sub(r' +', ' ', text)

    return text.strip()


def composition_percentages(cell_str):
    output = {"numbers": 0, "text": 0, "special": 0}

    # clean
    cell_str_cleaned = re.sub('[^a-zA-Z0-9]', '', cell_str)

    chars = len(cell_str_cleaned)

    if chars == 0:
        return output

    numbers = re.sub(r'[^0-9]', "", cell_str_cleaned)

    output['numbers'] = len(numbers) / chars

    text = re.sub(r'[^a-zA-Z]', "", cell_str_cleaned)

    output['text'] = len(text) / chars

    # for special chars, look at all characters
    special = re.sub(r'[a-zA-Z0-9 ]', "", cell_str)

    output['special'] = len(special) / len(cell_str)

    return output


def is_year_cell(cell_str):
    pieces = words_contained(cell_str)
    for p in pieces:
        if p in years_abs_strings:
            return True
    return False


def get_mean_for_column(column_values: List) -> float:
    values_filtered = [abs(x) for x in column_values if x is not None and is_number_cell(x)]
    if len(values_filtered) == 0:
        return 0
    return float(np.mean(values_filtered))


def clean_spaces(cell_str):
    return " ".join([x for x in cell_str.split(" ") if x != ""])


def delete_trailing_zeros(integer) -> int:
    string = str(integer).rstrip('0')
    return int(string) if string != '' else 0


def clean_number_for_matching(num) -> int:
    if num is None:
        return 0
    # clean comma, dots etc.
    num = re.sub(r'[^0-9]', '', str(num))

    # make abs
    num = abs(int(num))

    # delete trailing zeros
    num = delete_trailing_zeros(num)

    return num


def get_source_identifier(file_path: str) -> str:
    BUF_SIZE = 65536

    sha = hashlib.sha256()

    with open(file_path, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha.update(data)

    return sha.hexdigest()


from enum import Enum


class DocumentType(Enum):
    HTML = "html"
    PDF = "pdf"


class ElementType(Enum):
    TEXT = "text"
    TABLE = "table"
    FIGURE = "figure"


class OutputType(Enum):
    INTEGER = "int"
    LIST = "list"
    TEXT = "text"
    NUMERIC = "numeric"
    ENTITY = "entity"
    DATE = "date"
    PERCENTAGE = "percentage"


class ContextType(Enum):
    QUESTIONS = "questions"
    ITEMS = "items"


class SearchStrategy(Enum):
    VECTOR = "vector"
    START = "start"


class RunMethod(Enum):
    WORKER = "worker"
    API = "api"


class AggregationMethod(Enum):
    SUM = "SUM"
    MAX = "MAX"


class ModelType(Enum):
    GPT = "gpt"
    CUSTOM = "custom"
    REPLICATE = "replicate"



from decimal import Decimal
from typing import *

from tiktoken.core import Encoding

from parsee.extraction.models.llm_models.prompts import Prompt


def get_tokens_encoded(prompt: str, encoding: Encoding) -> List[int]:
    return encoding.encode(prompt)


def truncate_prompt(prompt: str, encoding: Encoding, max_tokens: int) -> Tuple[str, int]:
    tokens = get_tokens_encoded(prompt, encoding)
    num_tokens = len(tokens)
    if num_tokens > max_tokens:
        prompt = encoding.decode(tokens[0:max_tokens])
        num_tokens = max_tokens
    return prompt, num_tokens


class LLMBaseModel:

    model_name: str

    def __init__(self, name: str):
        self.model_name = name

    def make_prompt_request(self, prompt: str) -> Tuple[str, Decimal]:
        raise NotImplemented



from typing import *
from functools import reduce

from parsee.extraction.models.llm_models.prompts import Prompt
from parsee.extraction.models.llm_models.structuring_schema import get_prompt_schema_item
from parsee.templates.general_structuring_schema import GeneralQueryItemSchema, StructuringItemSchema
from parsee.extraction.tasks.questions.utils import MAIN_QUESTION_STR
from parsee.extraction.extractor_dataclasses import ParseeMeta, ExtractedSource
from parsee.utils.enums import DocumentType, SearchStrategy
from parsee.storage.interfaces import StorageManager
from parsee.extraction.extractor_elements import StandardDocumentFormat, ExtractedEl


class GeneralQueriesPromptBuilder:

    storage: StorageManager

    def __init__(self, storage: StorageManager):
        self.storage = storage

    def build_raw_value(self, main_value: any, meta: List[ParseeMeta], sources: List[ExtractedSource], item: GeneralQueryItemSchema, meta_items: List[StructuringItemSchema]):
        schema_item = get_prompt_schema_item(item)
        if len(meta_items) > 0:
            output = f"{MAIN_QUESTION_STR}: {main_value}"
            for meta_item in meta_items:
                meta_schema_item = get_prompt_schema_item(meta_item)
                meta_answers = [x for x in meta if x.class_id == meta_item.id]
                answer_chosen = meta_answers[0].class_value if len(meta_answers) > 0 else "n/a"
                output += f"\n({meta_item.id}): {meta_schema_item.parsed_to_raw(answer_chosen)}"
        else:
            output = f"{schema_item.parsed_to_raw(main_value)}"
        output += "\nSources: " + ",".join(f"[{x.element_index}]" for x in sources)
        return output

    def get_relevant_elements(self, schema_item: GeneralQueryItemSchema, document: StandardDocumentFormat) -> List[ExtractedEl]:

        if schema_item.searchStrategy == SearchStrategy.VECTOR:
            return self.storage.vector_store.find_closest_elements(document, schema_item.title, schema_item.keywords, False)
        elif schema_item.searchStrategy == SearchStrategy.START:
            return document.elements
        else:
            raise NotImplemented

    def get_elements_text(self, elements: List[ExtractedEl]):
        llm_text = " (For the following data, if tables have empty cells, they are omitted)\n"
        for el in elements:
            llm_text += f"[{el.source.element_index}]: {el.get_text_llm(True)}\n"
        return llm_text

    def build_prompt(self, structuring_item: GeneralQueryItemSchema, meta_items: List[StructuringItemSchema], document: StandardDocumentFormat, relevant_elements_custom: Optional[List[ExtractedEl]] = None) -> Prompt:

        elements = self.get_relevant_elements(structuring_item, document) if relevant_elements_custom is None else relevant_elements_custom

        general_info = "You are supposed to answer a question based on text fragments that are provided. " \
                       "The fragments start with a number and then the actual text. The lower the number of the fragment, " \
                       "the earlier the fragment appeared in the document (reading from left to right, top to bottom). " \
                       "Please respond as concise as possible."

        prompt_schema_item = get_prompt_schema_item(structuring_item)

        additional_info_str = f" Additional info: {structuring_item.additionalInfo}" if structuring_item.additionalInfo.strip() != "" else ""
        main_question = f'The question is: {structuring_item.title} {additional_info_str} {prompt_schema_item.get_possible_values_str()}'

        relevant_meta_items = [x for x in meta_items if structuring_item.metaInfoIds is not None and x.id in structuring_item.metaInfoIds]

        # build full example
        source_examples = [ExtractedSource(DocumentType.PDF, None, None, 241, None), ExtractedSource(DocumentType.PDF, None, None, 423, None)]
        meta_examples = [ParseeMeta("test", 0, source_examples, x.id, get_prompt_schema_item(x).get_example(True), 0.8) for x in relevant_meta_items]
        example_output = self.build_raw_value(prompt_schema_item.get_example(), meta_examples, source_examples, structuring_item, meta_items)

        full_example = f"Your answer could look like this: {example_output}" if len(relevant_meta_items) == 0 else f"One possible answer block could be: {example_output}"

        # add prompting for meta
        if len(relevant_meta_items) > 0:
            main_question += "\n We also want to retrieve some meta information. In the following we will present the meta item ID and then the additional question to be answered, format: (META_ID): QUESTION. In the answer please first provide the meta ID (for the main question use 'main question' instead) in brackets and then the answer. If there are several correct answers to the question that are differentiated by their meta values, structure your answer into multiple answer blocks of the same format, one after the other separated by new lines."
            for meta_item in relevant_meta_items:
                meta_prompt_item = get_prompt_schema_item(meta_item)
                main_question += f"\n({meta_item.id}): {meta_item.title} {meta_item.additionalInfo} {meta_prompt_item.get_possible_values_str()}"

        prompt = Prompt(general_info, main_question,
                        'Please at the end of each answer also provide the numbers of the text fragments you used to answer in square brackets.',
                        full_example,
                        self.get_elements_text(elements))

        return prompt


from typing import *
import os

from parsee.extraction.models.model_dataclasses import *
from parsee.utils.constants import NUM_TOKENS_DEFAULT_LLM


def gpt_config(openai_api_key: str, token_limit: Optional[int] = None, openai_model_name: Optional[str] = None) -> MlModelSpecification:
    model_name = openai_model_name if openai_model_name is not None else "gpt-4-0314"
    token_limit = NUM_TOKENS_DEFAULT_LLM if token_limit is None else token_limit
    return MlModelSpecification("GPT 4", model_name, ModelType.GPT, None, None, token_limit, openai_api_key, None, None, None, None, None)


def replicate_config(replicate_api_key: str, model_name: str, token_limit: Optional[int] = None) -> MlModelSpecification:
    os.environ["REPLICATE_API_TOKEN"] = replicate_api_key
    token_limit = NUM_TOKENS_DEFAULT_LLM if token_limit is None else token_limit
    return MlModelSpecification(f"Replicate model: {model_name}", model_name, ModelType.REPLICATE, None, None, token_limit, None, None, None, None, None, None)


import os
import uuid
from typing import *
import pickle
import io

import numpy as np
import faiss
from sentence_transformers import SentenceTransformer

from parsee.storage.vector_stores.interfaces import VectorStore
from parsee.extraction.extractor_elements import StandardDocumentFormat, ExtractedEl
from parsee.utils.enums import ElementType


class SimpleFaissStore(VectorStore):

    def __init__(self):
        self.encoder = SentenceTransformer('all-MiniLM-L6-v2')
        self.min_chunk_size_characters = 1000
        self.k = 100

    def make_index(self, document: StandardDocumentFormat, tables_only: bool) -> Tuple[List[List[int]], any]:
        data = []
        current_batch = {"chars": 0, "element_indices": [], "text": ""}
        for k, el in enumerate(document.elements):
            if el.el_type == ElementType.TABLE:
                data.append(
                    {"element_indices": [el.source.element_index], "text": el.get_text()}
                )
            elif el.el_type == ElementType.TEXT and not tables_only:
                current_text = el.get_text_llm(False)

                current_batch["chars"] += len(current_text)
                current_batch["element_indices"].append(el.source.element_index)
                current_batch["text"] += "\n " if current_batch["text"] != "" else ""
                current_batch["text"] += current_text

                if current_batch["chars"] > self.min_chunk_size_characters:
                    data.append(
                        {"element_indices": current_batch["element_indices"], "text": current_batch["text"]}
                    )
                    current_batch = {"chars": 0, "element_indices": [], "text": ""}
        # check last batch
        if current_batch["chars"] > 0:
            data.append(
                {"element_indices": current_batch["element_indices"], "text": current_batch["text"]}
            )

        sentence_embeddings = self.encoder.encode([x["text"] for x in data])

        d = sentence_embeddings.shape[1]
        index = faiss.IndexFlatL2(d)
        index.add(sentence_embeddings)

        return [x["element_indices"] for x in data], index

    def get_index(self, document: StandardDocumentFormat, tables_only: bool):
        return self.make_index(document, tables_only)

    def find_closest_elements(self, document: StandardDocumentFormat, search_element_title: str, keywords: Optional[str], tables_only: bool = True) -> List[ExtractedEl]:

        element_indices, index = self.get_index(document, tables_only)

        query = f"{search_element_title}" + (f"; {keywords}" if keywords is not None else "")

        xq = self.encoder.encode([query])

        _, results = index.search(xq, self.k)  # search

        indices = results.tolist()[0]

        all_element_indices = []
        for idx in indices:
            if idx >= 0:
                all_element_indices += element_indices[idx]

        return [document.elements[el_idx] for el_idx in all_element_indices]



from typing import *
from dataclasses import dataclass
from decimal import Decimal
import os

from parsee.utils.enums import ModelType


@dataclass
class MlModelSpecification:
    name: str
    internal_name: str
    model_type: ModelType
    file_path: Optional[str]
    price_per_1k_tokens: Optional[Decimal]
    max_tokens: Optional[int]
    api_key: Optional[str]
    only_questions: Optional[List[str]]
    only_elements: Optional[List[str]]
    only_meta: Optional[List[str]]
    only_mappings: Optional[List[str]]
    stats: Optional[Dict]

    def model_path(self) -> Union[None, str]:
        if self.file_path is None:
            return None
        return os.path.join(self.file_path, "best_model.h5")

    def settings_path(self) -> Union[None, str]:
        if self.file_path is None:
            return None
        return os.path.join(self.file_path, "settings.pkl")


import os
import time
from typing import List, Tuple
from dataclasses import dataclass
from decimal import Decimal
import math

import tiktoken
import replicate

from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel, get_tokens_encoded, truncate_prompt
from parsee.extraction.models.model_dataclasses import MlModelSpecification


class ReplicateModel(LLMBaseModel):

    def __init__(self, model: MlModelSpecification):
        super().__init__(model.name)
        self.model = model
        self.max_retries = 5
        self.encoding = tiktoken.get_encoding("cl100k_base")
        self.max_tokens_answer = 256
        self.max_requests_for_summarization = 30
        self.max_tokens_question = self.model.max_tokens - self.max_tokens_answer

    def slice_data(self, data: str) -> List[str]:
        tokens = get_tokens_encoded(data, self.encoding)
        slices = []
        num_slices = math.ceil(len(tokens) / self.max_tokens_question)
        for slice_num in range(0, num_slices):
            current_slice = self.encoding.decode(tokens[slice_num*self.max_tokens_question:slice_num*self.max_tokens_question+self.max_tokens_question])
            slices.append(current_slice)
        return slices

    def _call_api(self, prompt: str, retries: int = 0, wait: int = 5) -> str:
        try:
            response = replicate.run(self.model.internal_name, input={
                "prompt": prompt,
                "top_k": 50,
                "top_p": 0.9,
                "temperature": 0,
                "max_new_tokens": 1024,
                "presence_penalty": 0,
                "frequency_penalty": 0
            })
            answer = "".join(response)
            return answer
        except Exception as e:
            if retries < self.max_retries:
                time.sleep(wait * 2 ** retries)
                return self._call_api(prompt, retries + 1)
            else:
                return ""

    def make_prompt_request(self, prompt: str) -> Tuple[str, Decimal]:
        final_prompt, num_tokens_input = truncate_prompt(prompt, self.encoding, self.max_tokens_question)
        response = self._call_api(final_prompt)
        tokens_response = len(get_tokens_encoded(response, self.encoding))
        final_cost = Decimal((tokens_response + num_tokens_input) * (self.model.price_per_1k_tokens/1000)) if self.model.price_per_1k_tokens is not None else Decimal(0)
        return response, final_cost



from pydantic.dataclasses import dataclass
from typing import Dict, Optional


@dataclass
class WebhookSchema:
    url: str
    headers: Dict
    sendData: bool


@dataclass
class ExportSchema:
    webhook: Optional[WebhookSchema] = None



from typing import *

from parsee.extraction.extractor_dataclasses import ParseeAnswer
from parsee.datasets.readers.interfaces import DatasetReader
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.templates.job_template import JobTemplate
from parsee.extraction.models.model_loader import ModelLoader, LLMQuestionModel
from parsee.storage.interfaces import StorageManager
from parsee.storage.in_memory_storage import InMemoryStorageManager


class EvaluationResult:

    def __init__(self):
        self.answers = {}

    def add_answers(self, source_identifier: str, answers: List[ParseeAnswer], is_assigned: bool):
        if len(answers) == 0:
            return
        if source_identifier not in self.answers:
            self.answers[source_identifier] = {}
        key = "assigned" if is_assigned else answers[0].model
        if key not in self.answers[source_identifier]:
            self.answers[source_identifier][key] = {}
        obj = self.answers[source_identifier][key]
        for answer in answers:
            if answer.class_id not in obj:
                obj[answer.class_id] = {}
            if answer.unique_id() not in obj[answer.class_id]:
                obj[answer.class_id][answer.unique_id()] = answer

    def evaluate(self) -> Dict:
        scores_by_source = {}
        for source, sc_dict in self.answers.items():
            scores_by_source[source] = {}
            reference_item = sc_dict["assigned"]
            for model, by_class in sc_dict.items():
                scores_by_source[source][model] = {"error_log": []}
                # check that all questions have been answered
                scores_by_source[source][model]["completion"] = len(by_class.keys()) / len(reference_item.keys())
                # check each question
                score = 0
                score_meta = 0
                for class_id, assigned_values in reference_item.items():
                    if class_id not in by_class:
                        continue
                    predicted_values = by_class[class_id]
                    for meta_key, correct_value in assigned_values.items():
                        if meta_key in predicted_values:
                            if meta_key != "":
                                score_meta += 1
                            predicted_value = predicted_values[meta_key]
                            # check if values match
                            if correct_value.class_value == predicted_value.class_value:
                                score += 1
                            else:
                                scores_by_source[source][model]["error_log"].append({"doc": source, "class_id": class_id, "type": "main question", "expected": correct_value.class_value, "actual": predicted_value.class_value})
                        elif len(assigned_values.keys()) == 1 and len(predicted_values.keys()) == 1:
                            predicted_value = predicted_values[list(predicted_values.keys())[0]]
                            # meta mismatch
                            scores_by_source[source][model]["error_log"].append({"doc": source, "class_id": class_id, "type": "meta", "expected": correct_value.meta_key(), "actual": predicted_value.meta_key()})
                            # check if values match
                            if correct_value.class_value == predicted_value.class_value:
                                score += 1
                            else:
                                scores_by_source[source][model]["error_log"].append({"doc": source, "class_id": class_id, "type": "main question", "expected": correct_value.class_value, "actual": predicted_value.class_value})

                scores_by_source[source][model]["total_correct"] = score
                scores_by_source[source][model]["total_correct_meta_found"] = score_meta

        total_scores = {}
        for source, by_source in scores_by_source.items():
            for model, scores in by_source.items():
                if model not in total_scores:
                    total_scores[model] = {"completion": 0, "total_correct": 0, "total_correct_meta_found": 0, "error_log": []}
                for score_key, score_value in scores.items():
                    total_scores[model][score_key] += score_value
        reference_scores = total_scores["assigned"]
        for model, scores_dict in total_scores.items():
            scores_dict_final = {**scores_dict}
            for score_key, score_value in scores_dict.items():
                if score_key == "completion":
                    scores_dict_final[score_key] = scores_dict[score_key] / len(scores_by_source.keys())
                elif score_key == "error_log":
                    pass
                else:
                    rel_key = score_key+"_percent"
                    scores_dict_final[rel_key] = score_value / reference_scores[score_key]
            total_scores[model] = scores_dict_final
        return total_scores


def evaluate_llm_performance(template: JobTemplate, reader: DatasetReader, models: List[MlModelSpecification], storage: Optional[StorageManager] = None) -> Dict:

    storage = InMemoryStorageManager(models) if storage is None else storage
    loader = ModelLoader(storage)
    ev = EvaluationResult()

    for spec in models:
        if spec.model_type == "custom":
            raise Exception("custom models not allowed here")

    for row, _ in reader.row_generator():

        for k, model_spec in enumerate(models):
            model: LLMQuestionModel = loader.get_question_model(model_spec.internal_name, template.questions.items, template.meta, {})
            if model is None:
                raise Exception("model not found")
            schema_items = [x for x in template.questions.items if x.id == row.element_identifier]
            if len(schema_items) == 0:
                raise Exception("item not found in schema")
            schema_item = schema_items[0]
            answers_model = model.predict_for_prompt(row.get_feature("full_prompt"), schema_item, None, None)
            ev.add_answers(row.source_identifier, answers_model, False)
            if k == 0:
                answers_assigned = model.parse_prompt_answer(schema_item, row.get_value("answer", False), None, None)
                ev.add_answers(row.source_identifier, answers_assigned, True)

    return ev.evaluate()



from typing import *
from decimal import Decimal

from parsee.extraction.tasks.mappings.mapping_model import MappingModel, ElementSchema, MappingSchema, ParseeBucket
from parsee.storage.interfaces import StorageManager
from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel
from parsee.extraction.tasks.mappings.features import LLMMappingFeatureBuilder
from parsee.extraction.extractor_elements import FinalOutputTable
from parsee.utils.constants import ID_NOT_AVAILABLE


class MappingModelLLM(MappingModel):

    def __init__(self, items: List[ElementSchema], storage: StorageManager, llm: LLMBaseModel, **kwargs):
        super().__init__(items)
        self.model_name = llm.model_name
        self.llm = llm
        self.storage = storage
        self.prob = 0.8
        self.feature_builder: LLMMappingFeatureBuilder = LLMMappingFeatureBuilder()

    def parse_answer(self, answer: str, schema: MappingSchema, kv_idx: int, li_identifier: str) -> Union[None, ParseeBucket]:
        for bucket in schema.buckets:
            if self.feature_builder.item_id_string(bucket).lower() in answer.lower() or bucket.id.lower() in answer.lower():
                return ParseeBucket(schema.id, bucket.id, li_identifier, self.model_name, self.prob, kv_idx, 0, Decimal(1))
        return ParseeBucket(schema.id, ID_NOT_AVAILABLE, li_identifier, self.model_name, self.prob, kv_idx, 0, Decimal(1))

    def classify_with_schema(self, table: FinalOutputTable, schema: MappingSchema) -> List[ParseeBucket]:

        output = []
        for idx, _ in enumerate(table.line_items):
            prompt = self.feature_builder.make_prompt(table, schema, idx)
            answer, amount = self.llm.make_prompt_request(str(prompt))
            self.storage.log_expense(self.llm.model_name, amount, f"mapping:{table.detected_class}")
            output.append(self.parse_answer(answer, schema, idx, table.li_identifier))
        return output



from typing import *

from parsee.datasets.dataset_dataclasses import MappingUniqueIdentifier
from parsee.extraction.extractor_elements import FinalOutputTable
from parsee.extraction.extractor_dataclasses import ParseeBucket
from parsee.utils.helper import clean_text_for_word_vectors2
from parsee.extraction.models.llm_models.prompts import Prompt
from parsee.templates.mappings import MappingBucket
from parsee.datasets.dataset_dataclasses import DatasetRow
from parsee.extraction.tasks.mappings.mapping_model import MappingSchema


class MappingFeatureBuilder:

    _memory: Dict[MappingUniqueIdentifier, Dict]

    def __init__(self):
        self._memory = {}
        self.number_replacement = "xnumberx"

    def get_text_all(self, table: FinalOutputTable, schema_id: str) -> str:

        items = []
        for kv_index, item in enumerate(table.line_items):
            items.append(self._make_base_features(table, schema_id, kv_index)["caption_cleaned"])

        return ", ".join(items)

    def _get_cleaned_caption(self, table: FinalOutputTable, schema_id: str, kv_index: int) -> str:
        if kv_index < 0 or kv_index > len(table.line_items)-1:
            return ""
        features = self._make_base_features(table, schema_id, kv_index)
        return features["caption_cleaned"]

    def _make_base_features(self, table: FinalOutputTable, schema_id: str, kv_index: int) -> Dict[str, str]:

        unique_identifier = MappingUniqueIdentifier(schema_id, table.li_identifier, kv_index)

        if unique_identifier in self._memory:
            return self._memory[unique_identifier]

        line_item_org = table.line_items[kv_index]
        total_items = len(table.line_items)

        self._memory[unique_identifier] = {"class_id": table.detected_class, "caption_org": line_item_org, "caption_cleaned": clean_text_for_word_vectors2(line_item_org, None, True, True, self.number_replacement), "item_idx": kv_index, "total_items": total_items}

        return self._memory[unique_identifier]

    def make_features(self, source_identifier: Optional[str], template_id: Optional[str], table: FinalOutputTable, schema_id: str, kv_index: int) -> DatasetRow:

        base_features = self._make_base_features(table, schema_id, kv_index)
        unique_identifier = MappingUniqueIdentifier(schema_id, table.li_identifier, kv_index)

        all_text = self.get_text_all(table, schema_id)

        item_before = self._get_cleaned_caption(table, schema_id, kv_index-1)
        item_after = self._get_cleaned_caption(table, schema_id, kv_index+1)

        features = {**base_features, "all_items": all_text, "item_before": item_before, "item_after": item_after}

        return DatasetRow(source_identifier, template_id, unique_identifier, features)


class LLMMappingFeatureBuilder(MappingFeatureBuilder):

    def __init__(self):
        super().__init__()
        self.number_replacement = "[number]"

    def build_raw_answer(self, bucket_choice: ParseeBucket, schema: MappingSchema) -> str:
        mapping_bucket = [x for x in schema.buckets if x.id == bucket_choice.bucket_id]
        if len(mapping_bucket) == 0:
            return "n/a"
        return self.item_id_string(mapping_bucket[0])

    def item_id_string(self, bucket: MappingBucket):
        return f"[{bucket.id}]"

    def build_prompt(self, features: DatasetRow, schema: MappingSchema) -> Prompt:

        bucket_str = ""
        for schema_item in schema.buckets:
            bucket_str += f"{self.item_id_string(schema_item)}: {schema_item.caption}\n"
        return Prompt(
            "You are supposed to classify a line-item from a table into exactly one of several possible buckets.",
            f"The item is called: '{features.get_feature('caption_org')}' (index of item in table: {features.get_feature('item_idx')}). You are only supposed to classify this item.\n",
            f"{'' if features.get_feature('item_before') == '' else 'For reference, the item before the relevant item in the table is the following: '+features.get_feature('item_before')+'.'}" +
            f"{'' if features.get_feature('item_after') == '' else 'Also for reference, the item after the relevant item in the table is the following: '+features.get_feature('item_after')+'.'}",
            f"Each bucket has an ID, only use the ID of the bucket in your answer and nothing else. Your answer could be for example: {self.item_id_string(schema.buckets[0])}",
            f"The possible buckets are the following (each bucket name and description is preceded by the ID in square brackets): {bucket_str}"
        )

    def make_prompt(self, table: FinalOutputTable, schema: MappingSchema, kv_index: int) -> Prompt:

        features = self.make_features(None, None, table, schema.id, kv_index)

        return self.build_prompt(features, schema)


from typing import *

from parsee.extraction.extractor_elements import ExtractedEl, get_text_distance, StandardDocumentFormat
from parsee.extraction.extractor_dataclasses import ParseeLocation
from parsee.templates.element_schema import ElementSchema
from parsee.utils.helper import words_contained, clean_text_for_word_vectors2, composition_percentages
from parsee.utils.enums import ElementType, SearchStrategy
from parsee.datasets.dataset_dataclasses import DatasetRow
from parsee.utils.constants import ELEMENTS_WORDS_TO_INCLUDE, ELEMENTS_TABLES_TO_INCLUDE
from parsee.storage.interfaces import StorageManager
from parsee.extraction.models.llm_models.prompts import Prompt


class LocationFeatureBuilder:

    memory: Dict[int, Dict[str, any]]

    def __init__(self):
        self.memory = {}
        self.number_replacement = "xnumberx"

    def _get_base_features(self, element_index: int, element: ExtractedEl) -> Dict[str, any]:

        if element_index in self.memory:
            return self.memory[element_index]

        el_dict = {}
        el_text = element.get_text()
        num_words = len(words_contained(el_text))
        el_dict['type'] = element.el_type.value
        el_dict['text'] = element.get_text_llm(False)
        el_dict['text_clean'] = clean_text_for_word_vectors2(el_text, remove_special_chars=True, remove_all_numbers=True, number_token=self.number_replacement)
        el_dict['num_words'] = num_words
        el_dict["percent_numbers"] = composition_percentages(el_text)['numbers']

        self.memory[element_index] = el_dict
        return el_dict

    def make_features(self, source_identifier: Optional[str], template_id: Optional[str], element_indices: List[int], elements: List[ExtractedEl], include_tables: bool = True) -> List[DatasetRow]:

        output = []

        for el_idx in element_indices:

            base_features = self._get_base_features(el_idx, elements[el_idx])
            text_before = self._get_text_features(el_idx, True, elements, ELEMENTS_WORDS_TO_INCLUDE)
            text_after = self._get_text_features(el_idx, False, elements, ELEMENTS_WORDS_TO_INCLUDE)
            tables_before = self._get_table_features(el_idx, True, elements, ELEMENTS_TABLES_TO_INCLUDE if include_tables else 0)
            tables_after = self._get_table_features(el_idx, False, elements, ELEMENTS_TABLES_TO_INCLUDE if include_tables else 0)

            full_features = {
                **base_features,
                "text_before": text_before,
                "text_after": text_after,
                **tables_before,
                **tables_after
            }

            row = DatasetRow(source_identifier, template_id, el_idx, full_features)

            output.append(row)

        return output

    # returns forward/backward looking text features
    def _get_text_features(self, idx: int, backward: bool, elements: List[ExtractedEl], word_limit: int) -> str:

        text_entries_to_take = []
        word_counter = 0

        range_generator = range(idx + 1, len(elements)) if not backward else range(idx - 1, -1, -1)

        for kk in range_generator:

            element = elements[kk]

            if element.el_type == ElementType.TEXT:

                el_features = self._get_base_features(kk, elements[kk])

                word_counter += el_features['num_words']
                if backward:
                    text_entries_to_take.append(el_features['text_clean'])
                else:
                    text_entries_to_take.insert(0, el_features["text_clean"])

                if word_counter >= word_limit:
                    break

        # take zeros entry
        if len(text_entries_to_take) == 0:
            return ""
        else:
            return (" ".join(text_entries_to_take)).strip()

    # returns features for tables before/after actual table
    def _get_table_features(self, idx: int, backward: bool, elements: List[ExtractedEl], table_limit: int) -> Dict[str, any]:
    
        tables_added = 0
        temp_features = {}
    
        range_generator = range(idx + 1, len(elements)) if not backward else range(idx - 1, -1, -1)
    
        for kk in range_generator:

            element = elements[kk]

            if element.el_type == ElementType.TABLE:

                table_features = self._get_base_features(kk, element)

                tables_added += 1
    
                # determine text distance between tables
                text_distance = get_text_distance(element.source.element_index, elements[idx].source.element_index, elements)
    
                # add all features
                key_append = f"_before_{tables_added}" if backward else f"_after_{tables_added}"
                temp_features["text_distance"+key_append] = text_distance
                temp_features["percent_numbers" + key_append] = table_features["percent_numbers"]
                temp_features["num_words" + key_append] = table_features["num_words"]
                temp_features["text_clean" + key_append] = table_features["text_clean"]
                temp_features["text_simple" + key_append] = table_features["text"]
    
                if tables_added >= table_limit:
                    break
    
        # if not enough tables added, add empty features
        if tables_added < table_limit:
            for a in range(0, table_limit - tables_added):
                idx_offset = abs(a+1+tables_added)
                key_append = f"_before_{idx_offset}" if backward else f"_after_{idx_offset}"
                temp_features["text_distance" + key_append] = 0
                temp_features["percent_numbers" + key_append] = 0
                temp_features["num_words" + key_append] = 0
                temp_features["text_clean" + key_append] = ""
                temp_features["text_simple" + key_append] = ""
    
        return temp_features


class LLMLocationFeatureBuilder(LocationFeatureBuilder):

    def __init__(self):
        super().__init__()
        self.number_replacement = "[number]"

    def build_raw_answer(self, locations: List[ParseeLocation]) -> str:
        sorted_ids = list(sorted([x.source.element_index for x in locations]))
        return f"[{','.join([str(x) for x in sorted_ids])}]"

    def get_elements_text(self, features: List[DatasetRow]):
        llm_text = ""
        for feature_entry in features:
            llm_text += f"[{feature_entry.element_identifier}] text before table: {feature_entry.get_feature('text_before')}; line items in table: {feature_entry.get_feature('text')} [end of item] \n"
        return llm_text

    def make_prompt(self, item: ElementSchema, document: StandardDocumentFormat, storage: StorageManager) -> Prompt:

        if item.searchStrategy == SearchStrategy.VECTOR:
            closest_elements = storage.vector_store.find_closest_elements(document, item.title, item.keywords, True)
        elif item.searchStrategy == SearchStrategy.START:
            closest_elements = [el for el in document.elements if el.el_type == ElementType.TABLE]
        else:
            raise NotImplemented

        additional_info_str = f" Additional info: {item.additionalInfo}" if item.additionalInfo.strip() != "" else ""
        all_element_indices = [x.source.element_index for x in closest_elements]

        features = self.make_features(None, None, all_element_indices, document.elements, False)

        prompt = Prompt("", f'we want to find an item labeled "{item.title}".{additional_info_str}', f"""We are providing data in the following with the element_index and the text, such as [ELEMENT_INDEX] "TEXT" [end of item].
                Using the text, identify "{item.title}" and return the elements which you think are most likely representing "{item.title}" (it can be one item or several together).
                It is very important that you only return the element_index and nothing else.""", "Your response could be for example: [10] or [241],[1204] ", self.get_elements_text(features))

        return prompt
